{{component (concat "job-page/" model.templateType)
  job=model
  sortProperty=sortProperty
  sortDescending=sortDescending
  currentPage=currentPage
  onNamespaceChange=(action "gotoJobs")
  gotoJob=(action "gotoJob")
  gotoTaskGroup=(action "gotoTaskGroup")}}
