<div class="accordion-head-content">
  {{yield}}
</div>
<button
  data-test-accordion-toggle
  class="button is-light is-compact pull-right accordion-toggle {{unless isExpandable "is-invisible"}}"
  onclick={{action (if isOpen onClose onOpen) item}}>
  {{buttonLabel}}
</button>
