{{#global-header class="page-header"}}
  {{#each breadcrumbs as |breadcrumb index|}}
    <li class="{{if (eq (inc index) breadcrumbs.length) "is-active"}}">
      {{#link-to data-test-breadcrumb=breadcrumb.label params=breadcrumb.args}}{{breadcrumb.label}}{{/link-to}}
    </li>
  {{/each}}
{{/global-header}}
{{#job-page/parts/body job=job onNamespaceChange=onNamespaceChange}}
  {{job-page/parts/error errorMessage=errorMessage onDismiss=(action "clearErrorMessage")}}

  {{#job-page/parts/title job=job title=job.trimmedName handleError=(action "handleError")}}
    <span class="tag is-hollow">periodic</span>
    <button data-test-force-launch class="button is-warning is-small is-inline" onclick={{action "forceLaunch"}}>Force Launch</button>
  {{/job-page/parts/title}}

  <div class="boxed-section job-stats">
    <div class="boxed-section-body">
      <span data-test-job-stat="version"><strong>Version:</strong> {{job.version}} | </span>
      <span data-test-job-stat="priority"><strong>Priority:</strong> {{job.priority}} </span>
      {{#if (and job.namespace system.shouldShowNamespaces)}}
        <span data-test-job-stat="namespace"> | <strong>Namespace:</strong> {{job.namespace.name}}</span>
      {{/if}}
      <span data-test-job-stat="cron"> | <strong>Cron:</strong> <code>{{job.periodicDetails.Spec}}</code></span>
    </div>
  </div>

  {{job-page/parts/summary job=job}}
  {{job-page/parts/children
    job=job
    sortProperty=sortProperty
    sortDescending=sortDescending
    currentPage=currentPage
    gotoJob=gotoJob}}
{{/job-page/parts/body}}
