---
layout: "docs"
page_title: "Commands: namespace apply"
sidebar_current: "docs-commands-namespace-apply"
description: >
  The namespace apply command is used create or update a namespace.
---

# Command: namespace apply

The `namespace apply` command is used create or update a namespace.

~> Namespace commands are new in Nomad 0.7 and are only available with Nomad
Enterprise.

## Usage

```
nomad namespace apply [options] <namespace>
```

The `namespace apply` command requires the name of the namespace to be created
or updated.

## General Options

<%= partial "docs/commands/_general_options" %>

## Apply Options

* `-quota` : An optional quota to apply to the namespace.

* `-description` : An optional human readable description for the namespace.

## Examples

Create a namespace with a quota

```
$ nomad namespace apply -description "Prod API servers" -quota prod api-prod 
Successfully applied namespace "api-prod"!
```
