---
layout: "docs"
page_title: "Commands: quota inspect"
sidebar_current: "docs-commands-quota-inspect"
description: >
  The quota inspect command is used to view raw information about a particular
  quota specification.
---

# Command: quota inspect

The `quota inspect` command is used to view raw information about a particular
quota.

~> Quota commands are new in Nomad 0.7 and are only available with Nomad
Enterprise.

## Usage

```
nomad quota inspect [options] <quota_name>
```

## General Options

<%= partial "docs/commands/_general_options" %>

## Inspect Options

* `-t` : Format and display the quota using a Go template.

## Examples

Inspect a quota specification:

```
$ nomad quota inspect default-quota
{
    "Spec": {
        "CreateIndex": 8,
        "Description": "Limit the shared default namespace",
        "Limits": [
            {
                "Hash": "NLOoV2WBU8ieJIrYXXx8NRb5C2xU61pVVWRDLEIMxlU=",
                "Region": "global",
                "RegionLimit": {
                    "CPU": 2500,
                    "DiskMB": 0,
                    "IOPS": 0,
                    "MemoryMB": 2000,
                    "Networks": null
                }
            }
        ],
        "ModifyIndex": 56,
        "Name": "default-quota"
    },
    "UsageLookupErrors": {},
    "Usages": {
        "global": {
            "CreateIndex": 8,
            "ModifyIndex": 56,
            "Name": "default-quota",
            "Used": {
                "NLOoV2WBU8ieJIrYXXx8NRb5C2xU61pVVWRDLEIMxlU=": {
                    "Hash": "NLOoV2WBU8ieJIrYXXx8NRb5C2xU61pVVWRDLEIMxlU=",
                    "Region": "global",
                    "RegionLimit": {
                        "CPU": 500,
                        "DiskMB": 0,
                        "IOPS": 0,
                        "MemoryMB": 256,
                        "Networks": null
                    }
                }
            }
        }
    }
}
```
