<div data-test-error class="empty-message">
  <h3 data-test-error-title class="empty-message-headline">Not Authorized</h3>
  <p data-test-error-message class="empty-message-body">
    {{#if token.secret}}
      Your {{#link-to "settings.tokens"}}ACL token{{/link-to}} does not provide the required permissions. Contact your administrator if this is an error.
    {{else}}
      Provide an {{#link-to "settings.tokens"}}ACL token{{/link-to}} with requisite permissions to view this.
    {{/if}}
  </p>
</div>
