---
layout: "docs"
page_title: "Commands: operator keygen"
sidebar_current: "docs-commands-operator-keygen"
description: >
  The `operator keygen` command generates an encryption key that can be used for Nomad
  server's gossip traffic encryption. The keygen command uses a
  cryptographically strong pseudo-random number generator to generate the key.
---

# Command: operator keygen

The `operator keygen` command generates an encryption key that can be used for Nomad
server's gossip traffic encryption.  The keygen command uses a cryptographically
strong pseudo-random number generator to generate the key.

## Usage 

```
nomad operator keygen
```

## Example

```
nomad operator keygen
YgZOXLMhC7TtZqeghMT8+w==
```

