.chart.distribution-bar {
  display: block;
  height: 100%;

  svg {
    display: inline-block;
    height: 100%;
    width: 100%;

    .bars {
      rect {
        transition: opacity 0.3s ease-in-out;
        opacity: 1;
      }

      .inactive {
        opacity: 0.2;
      }

      .target {
        opacity: 0;
      }

      $color-sequence: $orange, $yellow, $green, $turquoise, $blue, $purple, $red;

      @for $i from 1 through length($color-sequence) {
        .slice-#{$i - 1} {
          fill: nth($color-sequence, $i);
        }
      }
    }
  }

  &.split-view {
    display: flex;
    flex-direction: row;
    align-items: center;

    svg {
      width: 50%;
      height: 30px;
    }

    .legend {
      list-style: none;
      width: 50%;
      padding: 1.5em;
      display: flex;
      flex-direction: row;
      flex-wrap: wrap;
      align-items: center;
      justify-content: center;

      li {
        display: block;
        background-color: transparent;
        transition: background-color 0.1s ease-in-out;
        border: 1px solid $grey-blue;
        padding: 0.25em 0.75em;
        margin: 0.25em;
        border-radius: $radius;

        // Ensure two columns, but don't use the full width
        width: 35%;

        .label,
        .value {
          display: inline;
          font-weight: $weight-normal;
        }

        &.is-active {
          background-color: rgba($info, 0.1);
        }

        &.is-empty {
          color: darken($grey-blue, 20%);
          border: none;

          .label {
            color: darken($grey-blue, 20%);
          }
        }
      }
    }
  }
}
