---
layout: "docs"
page_title: "Commands: acl token delete"
sidebar_current: "docs-commands-acl-token-delete"
description: >
  The token create command is used to delete existing ACL tokens.
---

# Command: acl token delete

The `acl token delete` command is used to delete existing ACL tokens.

## Usage

```
nomad acl token delete <token_accessor_id>
```

The `acl token delete` command requires an existing token's AccessorID.

## General Options

<%= partial "docs/commands/_general_options" %>

## Examples

Delete an existing ACL token:

```
$ nomad acl token delete d532c40a-30f1-695c-19e5-c35b882b0efd

Token d532c40a-30f1-695c-19e5-c35b882b0efd successfully deleted
```
