---
layout: "docs"
page_title: "Commands: alloc status"
sidebar_current: "docs-commands-alloc-status"
description: >
  Display status and metadata about existing allocations and their tasks.
---

# Command: alloc status

The `alloc status` command displays status information and metadata about an
existing allocation and its tasks. It can be useful while debugging to reveal
the underlying reasons for scheduling decisions or failures, as well as the
current state of its tasks. As of Nomad 0.7.1, alloc status also shows allocation
modification time in addition to create time. As of Nomad 0.8, alloc status shows
information about reschedule attempts.

## Usage

```
nomad alloc status [options] <allocation>
```

An allocation ID or prefix must be provided. If there is an exact match, the
full details of the allocation will be displayed. Otherwise, a list of matching
allocations and information will be displayed.

## General Options

<%= partial "docs/commands/_general_options" %>

## Alloc Status Options

* `-short`: Display short output. Shows only the most recent task event.
* `-verbose`: Show full information.
* `-json` : Output the allocation in its JSON format.
* `-t` : Format and display the allocation using a Go template.

## Examples

Short status of an alloc:

```
$ nomad alloc status --short 0af996ed
ID                  = 0af996ed
Eval ID             = be9bde98
Name                = example.cache[0]
Node ID             = 43c0b14e
Job ID              = example
Job Version         = 0
Client Status       = running
Client Description  = <none>
Desired Status      = run
Desired Description = <none>
Created At          = 07/25/17 16:12:48 UTC
Deployment ID       = 0c83a3b1
Deployment Health   = healthy

Tasks
Name   State    Last Event  Time
redis  running  Started     07/25/17 16:12:48 UTC
web    running  Started     07/25/17 16:12:49 UTC
```

Full status of an alloc, which shows one of the tasks dying and then being restarted:

```
$ nomad alloc status 0af996ed
ID                   = 0af996ed
Eval ID              = be9bde98
Name                 = example.cache[0]
Node ID              = 43c0b14e
Job ID               = example
Job Version          = 0
Client Status        = running
Client Description   = <none>
Desired Status       = run
Desired Description  = <none>
Created              = 5m ago
Modified             = 5m ago
Deployment ID        = 0c83a3b1
Deployment Health    = healthy
Replacement Alloc ID = 0bc894ca
Reschedule Attempts  = 1/3

Task "redis" is "running"
Task Resources
CPU        Memory           Disk     IOPS  Addresses
1/500 MHz  6.3 MiB/256 MiB  300 MiB  0     db: 127.0.0.1:27908

Task Events:
Started At     = 07/25/17 16:12:48 UTC
Finished At    = N/A
Total Restarts = 0
Last Restart   = N/A

Recent Events:
Time                   Type        Description
07/25/17 16:12:48 UTC  Started     Task started by client
07/25/17 16:12:48 UTC  Task Setup  Building Task Directory
07/25/17 16:12:48 UTC  Received    Task received by client

Task "web" is "running"
Task Resources
CPU        Memory           Disk     IOPS  Addresses
1/500 MHz  6.3 MiB/256 MiB  300 MiB  0     db: 127.0.0.1:30572

Task Events:
Started At     = 07/25/17 16:12:49 UTC
Finished At    = N/A
Total Restarts = 0
Last Restart   = N/A

Recent Events:
07/25/17 16:12:49 UTC  Started     Task started by client
07/25/17 16:12:48 UTC  Task Setup  Building Task Directory
07/25/17 16:12:48 UTC  Received    Task received by client
```

Verbose status can also be accessed:

```
$ nomad alloc status -verbose 0af996ed
ID                     = 0af996ed-aff4-8ddb-a566-e55ebf8969c9
Eval ID                = be9bde98-0490-1beb-ced0-012d10ddf22e
Name                   = example.cache[0]
Node ID                = 43c0b14e-7f96-e432-a7da-06605257ce0c
Job ID                 = example
Job Version            = 0
Client Status          = running
Client Description     = <none>
Desired Status         = run
Desired Description    = <none>
Created                = 07/25/17 16:12:48 UTC
Modified               = 07/25/17 16:12:48 UTC
Deployment ID          = 0c83a3b1-8a7b-136b-0e11-8383dc6c9276
Deployment Health      = healthy
Reschedule Eligibility = 2m from now
Evaluated Nodes        = 1
Filtered Nodes         = 0
Exhausted Nodes        = 0
Allocation Time        = 38.474µs
Failures               = 0

Task "redis" is "running"
Task Resources
CPU        Memory           Disk     IOPS  Addresses
1/500 MHz  6.3 MiB/256 MiB  300 MiB  0     db: 127.0.0.1:27908

Task Events:
Started At     = 07/25/17 16:12:48 UTC
Finished At    = N/A
Total Restarts = 0
Last Restart   = N/A

Recent Events:
Time                   Type        Description
07/25/17 16:12:48 UTC  Started     Task started by client
07/25/17 16:12:48 UTC  Task Setup  Building Task Directory
07/25/17 16:12:48 UTC  Received    Task received by client

Task "web" is "running"
Task Resources
CPU        Memory           Disk     IOPS  Addresses
1/500 MHz  6.3 MiB/256 MiB  300 MiB  0     db: 127.0.0.1:30572

Task Events:
Started At     = 07/25/17 16:12:49 UTC
Finished At    = N/A
Total Restarts = 0
Last Restart   = N/A

Recent Events:
Time                   Type        Description
07/25/17 16:12:49 UTC  Started     Task started by client
07/25/17 16:12:48 UTC  Task Setup  Building Task Directory
07/25/17 16:12:48 UTC  Received    Task received by client
```
