---
layout: "docs"
page_title: "Commands: quota list"
sidebar_current: "docs-commands-quota-list"
description: >
  The quota list command is used to list available quota specifications.
---

# Command: quota list

The `quota list` command is used to list available quota specifications.

~> Quota commands are new in Nomad 0.7 and are only available with Nomad
Enterprise.

## Usage

```
nomad quota list
```

## General Options

<%= partial "docs/commands/_general_options" %>

## List Options

* `-json`: Output the quota specifications in a JSON format.

* `-t`: Format and display the quotas specifications using a Go template.

## Examples

List all quota specifications:

```
$ nomad quota list
Name     Description
default  Limit the shared default namespace
```
