---
layout: "docs"
page_title: "Commands: namespace list"
sidebar_current: "docs-commands-namespace-list"
description: >
  The namespace list command is used to list namespaces.
---

# Command: namespace list

The `namespace list` command is used list available namespaces.

~> Namespace commands are new in Nomad 0.7 and are only available with Nomad
Enterprise.

## Usage

```
nomad namespace list [options]
```

The `namespace list` command requires no arguments.

## General Options

<%= partial "docs/commands/_general_options" %>

## List Options

* `-json` : Output the namespaces in their JSON format.

* `-t` : Format and display the namespaces using a Go template.

## Examples

List all namespaces:

```
$ nomad namespace list
Name      Description
default   Default shared namespace
api-prod  Production instances of backend API servers
api-qa    QA instances of backend API servers
web-prod  Production instances of webservers
web-qa    QA instances of webservers
```
