---
layout: "docs"
page_title: "Commands: acl policy info"
sidebar_current: "docs-commands-acl-policy-info"
description: >
  The policy info command is used to fetch information on an existing ACL
  policy.
---

# Command: acl policy info

The `acl policy info` command is used to fetch information on an existing ACL
policy.

## Usage

```
nomad acl policy info <name>
```

The `acl policy info` command requires the policy name.

## General Options

<%= partial "docs/commands/_general_options" %>

## Examples

Fetch information on an existing ACL Policy:

```
$ nomad acl policy info my-policy
Name        = my-policy
Description = <none>
Rules       = {
  "Name": "my-policy",
  "Description": "This is a great policy",
  "Rules": "list_jobs"
}
CreateIndex = 749
ModifyIndex = 758
```
