---
layout: "docs"
page_title: "Commands: quota delete"
sidebar_current: "docs-commands-quota-delete"
description: >
  The quota delete command is used to delete an existing quota specification.
---

# Command: quota delete

The `quota delete` command is used to delete an existing quota specification.

~> Quota commands are new in Nomad 0.7 and are only available with Nomad
Enterprise.

## Usage

```
nomad quota delete <quota_name>
```

The `quota delete` command requires the quota specification name as an argument.

## General Options

<%= partial "docs/commands/_general_options" %>

## Examples

Delete a quota specification:

```
$ nomad quota delete my-quota
Successfully deleted quota "my-quota"!
```
