---
layout: "docs"
page_title: "Commands: acl token info"
sidebar_current: "docs-commands-acl-token-info"
description: >
  The token info command is used to fetch information about an existing ACL token.
---

# Command: acl token info

The `acl token info` command is used to fetch information about an existing ACL token.

## Usage

```
nomad acl token info <token_accessor_id>
```

The `acl token info` command requires an existing token's AccessorID.

## General Options

<%= partial "docs/commands/_general_options" %>

## Examples

Fetch information about an existing ACL token:

```
$ nomad acl token info d532c40a-30f1-695c-19e5-c35b882b0efd
Accessor ID  = d532c40a-30f1-695c-19e5-c35b882b0efd
Secret ID    = 85310d07-9afa-ef53-0933-0c043cd673c7
Name         = my token
Type         = client
Global       = false
Policies     = [foo bar]
Create Time  = 2017-09-15 05:04:41.814954949 +0000 UTC
Create Index = 8
Modify Index = 8
```
