---
layout: "docs"
page_title: "Commands: ui"
sidebar_current: "docs-commands-ui"
description: >
  The ui command is used to open the Nomad Web UI.
---

# Command: ui

The `ui` command is used to open the Nomad Web UI.

## Usage

```
nomad ui [options] <identifier>
```

The `ui` command can be called with no arguments, in which case the UI homepage
will be opened in the default browser.

An identifier may be provided, in which case the UI will be opened to view the
details for that object. Supported identifiers are jobs, allocations and nodes.


## General Options

<%= partial "docs/commands/_general_options" %>

## Examples

Open the UI homepage:

```
$ nomad ui
Opening URL "http://127.0.0.1:4646"
```

Open the UI directly to look at a job:

```
$ nomad ui redis-job
http://127.0.0.1:4646/ui/jobs/redis-job
```

Open the UI directly to look at an allocation:

```
$ nomad ui d4005969
Opening URL "http://127.0.0.1:4646/ui/allocations/d4005969-b16f-10eb-4fe1-a5374986083d"
```
