<div data-test-deployment-task-groups class="boxed-section">
  <div class="boxed-section-head">
    Task Groups
  </div>
  <div class="boxed-section-body is-full-bleed">
    {{#list-table
      data-test-deployment-task-groups
      source=deployment.taskGroupSummaries
      class="task-groups" as |t|}}
      {{#t.head}}
        <th>Name</th>
        <th>Needs Promotion?</th>
        <th>Auto Revert?</th>
        <th>Canaries</th>
        <th>Allocs</th>
        <th>Healthy Allocs</th>
        <th>Unhealthy Allocs</th>
        {{#if inProgress}}
          <th>Progress Deadline</th>
        {{/if}}
      {{/t.head}}
      {{#t.body as |row|}}
        <tr data-test-deployment-task-group>
          <td data-test-deployment-task-group-name>{{row.model.name}}</td>
          <td data-test-deployment-task-group-promotion>
            {{#if row.model.requiresPromotion}}
              {{if row.model.promoted "No" "Yes"}}
            {{else}}
              N/A
            {{/if}}
          </td>
          <td data-test-deployment-task-group-auto-revert>{{if row.model.autoRevert "Yes" "No"}}</td>
          <td data-test-deployment-task-group-canaries>{{or row.model.placedCanaries 0}} / {{row.model.desiredCanaries}}</td>
          <td data-test-deployment-task-group-allocs>{{row.model.placedAllocs}} / {{row.model.desiredTotal}}</td>
          <td data-test-deployment-task-group-healthy>{{row.model.healthyAllocs}}</td>
          <td data-test-deployment-task-group-unhealthy>{{row.model.unhealthyAllocs}}</td>
          {{#if inProgress}}
            <td data-test-deployment-task-group-progress-deadline>
              <span class="nowrap">{{moment-format row.model.requireProgressBy "MM/DD/YY HH:mm:ss"}}</span>
            </td>
          {{/if}}
        </tr>
      {{/t.body}}
    {{/list-table}}
  </div>
</div>

