---
layout: "docs"
page_title: "Commands: deployment"
sidebar_current: "docs-commands-deployment"
description: >
  The deployment command is used to interact with deployments.
---

# Command: deployment
 
The `deployment` command is used to interact with deployments.

## Usage

Usage: `nomad deployment <subcommand> [options]`

Run `nomad deployment <subcommand> -h` for help on that subcommand. The following
subcommands are available:

* [`deployment fail`][fail] - Manually fail a deployment
* [`deployment list`][list] - List all deployments
* [`deployment pause`][pause] - Pause a deployment
* [`deployment promote`][promote] - Promote canaries in a deployment
* [`deployment resume`][resume] - Resume a paused deployment
* [`deployment status`][status] - Display the status of a deployment

[fail]: /docs/commands/deployment/fail.html "Manually fail a deployment"
[list]: /docs/commands/deployment/list.html "List all deployments"
[pause]: /docs/commands/deployment/pause.html "Pause a deployment"
[promote]: /docs/commands/deployment/promote.html "Promote canaries in a deployment"
[resume]: /docs/commands/deployment/resume.html "Resume a paused deployment"
[status]: /docs/commands/deployment/status.html "Display the status of a deployment"
