---
layout: "docs"
page_title: "Commands: alloc"
sidebar_current: "docs-commands-alloc"
description: >
  The alloc command is used to interact with allocations.
---

# Command: alloc
 
The `alloc` command is used to interact with allocations.

## Usage

Usage: `nomad alloc <subcommand> [options]`

Run `nomad alloc <subcommand> -h` for help on that subcommand. The following
subcommands are available:

* [`alloc fs`][fs] - Inspect the contents of an allocation directory
* [`alloc logs`][logs] - Streams the logs of a task
* [`alloc status`][status] - Display allocation status information and metadata

[fs]: /docs/commands/alloc/fs.html "Inspect the contents of an allocation directory"
[logs]: /docs/commands/alloc/logs.html "Streams the logs of a task"
[status]: /docs/commands/alloc/status.html "Display allocation status information and metadata"
