{{yield (hash
  metrics=(component "job-deployment/deployment-metrics" deployment=deployment)
  taskGroups=(component "job-deployment/task-groups" deployment=deployment)
  allocations=(component "job-deployment/deployment-allocations" deployment=deployment)
)}}
