<svg>
  <defs>
    <clipPath>
      <rect class="mask" x="0" y="0" width="100%" height="100%" rx="2px" ry="2px" />
    </clipPath>
  </defs>
  <g class="bars"></g>
</svg>
{{#if hasBlock}}
  {{yield (hash
    data=_data
    activeDatum=activeDatum
  )}}
{{else}}
  <div class="chart-tooltip {{if isActive "active" "inactive"}}" style={{tooltipStyle}}>
    <ol>
      {{#each _data as |datum index|}}
        <li class="{{if (eq datum.index activeDatum.index) "active"}}">
          <span class="label {{if (eq datum.value 0) "is-empty"}}">
            <span class="color-swatch {{if datum.className datum.className (concat "swatch-" index)}}" />
            {{datum.label}}
          </span>
          <span class="value">{{datum.value}}</span>
        </li>
      {{/each}}
    </ol>
  </div>
{{/if}}
