<td data-test-job-name>{{#link-to "jobs.job" job.plainId class="is-primary"}}{{job.name}}{{/link-to}}</td>
<td data-test-job-status>
  <span class="tag {{job.statusClass}}">{{job.status}}</span>
</td>
<td data-test-job-type>{{job.displayType}}</td>
<td data-test-job-priority>{{job.priority}}</td>
<td data-test-job-task-groups>
  {{#if job.taskGroupCount}}
    {{job.taskGroupCount}}
  {{else}}
    --
  {{/if}}
</td>
<td data-test-job-allocations>
  <div class="inline-chart">
    {{#if job.hasChildren}}
      {{#if (gt job.totalChildren 0)}}
        {{children-status-bar job=job isNarrow=true}}
      {{else}}
        <em class="is-faded">No Children</em>
      {{/if}}
    {{else}}
      {{allocation-status-bar allocationContainer=job isNarrow=true}}
    {{/if}}
  </div>
</td>
