---
layout: "docs"
page_title: "Commands: server members"
sidebar_current: "docs-commands-server-members"
description: >
  The server members command is used to display a list of the known server members and their status.
---

# Command: server members

The `server members` command displays a list of the known servers in the cluster
and their current status. Member information is provided by the gossip protocol,
which is only run on server nodes.

## Usage

```
nomad server members [options]
```

## General Options

<%= partial "docs/commands/_general_options" %>

## Server Members Options

* `-detailed`: Dump the basic member information as well as the raw set of tags
  for each member. This mode reveals additional information not displayed in the
  standard output format.

## Examples

Default view:

```
$ nomad server members
Name          Addr      Port  Status  Proto  Build     DC   Region
node1.global  10.0.0.8  4648  alive   2      0.1.0dev  dc1  global
node2.global  10.0.0.9  4648  alive   2      0.1.0dev  dc1  global
```

Detailed view:

```
$ nomad server members -detailed
Name   Addr      Port  Tags
node1  10.0.0.8  4648  bootstrap=1,build=0.1.0dev,vsn=1,vsn_max=1,dc=dc1,port=4647,region=global,role=nomad,vsn_min=1
node2  10.0.0.9  4648  bootstrap=0,build=0.1.0dev,vsn=1,vsn_max=1,dc=dc1,port=4647,region=global,role=nomad,vsn_min=1
```
