<td data-test-icon class="is-narrow">
  {{#if node.unhealthyDrivers.length}}
    <span class="tooltip text-center" aria-label="Client has unhealthy drivers">
      {{x-icon "warning" class="is-warning"}}
    </span>
  {{/if}}
</td>
<td data-test-client-id>{{#link-to "clients.client" node.id class="is-primary"}}{{node.shortId}}{{/link-to}}</td>
<td data-test-client-name class="is-200px is-truncatable" title="{{node.name}}">{{node.name}}</td>
<td data-test-client-status>{{node.status}}</td>
<td data-test-client-drain>
  {{#if node.isDraining}}
    <span class="status-text is-info">true</span>
  {{else}}
    false
  {{/if}}
</td>
<td data-test-client-eligibility>
  {{#if node.isEligible}}
    {{node.schedulingEligibility}}
  {{else}}
    <span class="status-text is-warning">{{node.schedulingEligibility}}</span>
  {{/if}}
</td>
<td data-test-client-address>{{node.httpAddr}}</td>
<td data-test-client-datacenter>{{node.datacenter}}</td>
<td data-test-client-allocations>
  {{#if node.allocations.isPending}}
    ...
  {{else}}
    {{node.allocations.length}}
  {{/if}}
</td>
