---
layout: "docs"
page_title: "Commands: sentinel list"
sidebar_current: "docs-commands-sentinel-list"
description: >
  The sentinel list command is used to list all installed Sentinel policies.
---

# Command: sentinel list

The `sentinel list` command is used to display all the installed Sentinel policies.

~> Sentinel commands are new in Nomad 0.7 and are only available with Nomad
Enterprise.

## Usage

```
nomad sentinel list [options]
```

The `sentinel list` command requires no arguments.

## General Options

<%= partial "docs/commands/_general_options" %>

## Examples

List all policies:

```
$ nomad sentinel list
Name  Scope       Enforcement Level  Description
foo   submit-job  advisory           my test policy
```
