<div class="boxed-section-head">
  Evaluations
</div>
<div class="boxed-section-body {{if sortedEvaluations.length "is-full-bleed"}} evaluations">
  {{#if sortedEvaluations.length}}
    {{#list-table source=sortedEvaluations as |t|}}
      {{#t.head}}
        <th>ID</th>
        <th>Priority</th>
        <th>Triggered By</th>
        <th>Status</th>
        <th>Placement Failures</th>
      {{/t.head}}
      {{#t.body as |row|}}
        <tr data-test-evaluation="{{row.model.shortId}}">
          <td data-test-id>{{row.model.shortId}}</td>
          <td data-test-priority>{{row.model.priority}}</td>
          <td data-test-triggered-by>{{row.model.triggeredBy}}</td>
          <td data-test-status>{{row.model.status}}</td>
          <td data-test-blocked>
            {{#if (eq row.model.status "blocked")}}
              N/A - In Progress
            {{else if row.model.hasPlacementFailures}}
              True
            {{else}}
              False
            {{/if}}
          </td>
        </tr>
      {{/t.body}}
    {{/list-table}}
  {{else}}
    <div data-test-empty-evaluations-list class="empty-message">
      <h3 data-test-empty-evaluations-list-headline class="empty-message-headline">No Evaluations</h3>
      <p class="empty-message-body">This is most likely due to garbage collection.</p>
    </div>
  {{/if}}
</div>
