<div class="page-layout">
  {{#global-header class="page-header"}}
  {{/global-header}}
  {{#gutter-menu class="page-body"}}
    <section class="section has-text-centered">{{partial "partials/loading-spinner"}}</section>
  {{/gutter-menu}}
</div>
