---
layout: "docs"
page_title: "Commands: acl token update"
sidebar_current: "docs-commands-acl-token-update"
description: >
  The token update command is used to update existing ACL tokens.
---

# Command: acl token update

The `acl token update` command is used to update existing ACL tokens.

## Usage

```
nomad acl token update [options] <token_accessor_id>
```

The `acl token update` command requires an existing token's accessor ID.

## General Options

<%= partial "docs/commands/_general_options" %>

## Create Options

* `-name`: Sets the human readable name for the ACL token.

* `-type`: Sets the type of token. Must be one of "client" (default), or "management".

* `-global`: Toggles the global mode of the token. Global tokens are replicated to all regions. Defaults false.

* `-policy`: Specifies a policy to associate with the token. Can be specified multiple times,
    but only with client type tokens.

## Examples

Update an existing ACL token:

```
$ nomad acl token update -name="my updated token" -policy=foo -policy=bar d532c40a-30f1-695c-19e5-c35b882b0efd
Accessor ID  = d532c40a-30f1-695c-19e5-c35b882b0efd
Secret ID    = 85310d07-9afa-ef53-0933-0c043cd673c7
Name         = my updated token
Type         = client
Global       = false
Policies     = [foo bar]
Create Time  = 2017-09-15 05:04:41.814954949 +0000 UTC
Create Index = 8
Modify Index = 8
```
