{{#global-header class="page-header"}}
  <li class="is-active">
    {{#link-to "jobs.index"}}Jobs{{/link-to}}
  </li>
{{/global-header}}
{{#gutter-menu class="page-body"}}
  <section class="section has-text-centered">{{partial "partials/loading-spinner"}}</section>
{{/gutter-menu}}
