---
layout: "docs"
page_title: "Commands: acl token self"
sidebar_current: "docs-commands-acl-token-self"
description: >
  The token self command is used to fetch information about the currently set
  ACL token.
---

# Command: acl token self

The `acl token self` command is used to fetch information about the currently set ACL token.

## Usage

```
nomad acl token self
```

## General Options

<%= partial "docs/commands/_general_options" %>

## Examples

Fetch information about an existing ACL token:

```
$ export NOMAD_TOKEN=d532c40a-30f1-695c-19e5-c35b882b0efd

$ nomad acl tokenjself
Accessor ID  = d532c40a-30f1-695c-19e5-c35b882b0efd
Secret ID    = 85310d07-9afa-ef53-0933-0c043cd673c7
Name         = my token
Type         = client
Global       = false
Policies     = [foo bar]
Create Time  = 2017-09-15 05:04:41.814954949 +0000 UTC
Create Index = 8
Modify Index = 8
```
