---
layout: "docs"
page_title: "Commands: namespace inspect"
sidebar_current: "docs-commands-namespace-inspect"
description: >
  The namespace inspect command is used to view raw information about a particular
  namespace.
---

# Command: namespace inspect

The `namespace inspect` command is used to view raw information about a particular
namespace.

~> Namespace commands are new in Nomad 0.7 and are only available with Nomad
Enterprise.

## Usage

```
nomad namespace inspect [options] <namespace_name>
```

## General Options

<%= partial "docs/commands/_general_options" %>

## Inspect Options

* `-t` : Format and display the namespace using a Go template.

## Examples

Inspect a namespace:

```
$ nomad namespace inspect default
{
    "CreateIndex": 5,
    "Description": "Default shared namespace",
    "ModifyIndex": 38,
    "Name": "default",
    "Quota": "shared-default-quota"
}
```
