.message {
  background: $body-background-color;

  .message-header {
    background: $body-background-color;
    color: $text;
    font-size: $size-5;
    font-weight: $weight-semibold;
    padding-left: 0;
    padding-right: 0;
  }

  .message-body {
    border: 1px solid $grey-lighter;
    border-radius: $radius;

    &.is-full-bleed {
      padding: 0;
    }

    .table {
      margin-bottom: 0;
    }
  }
}
