---
layout: "docs"
page_title: "Commands: operator autopilot get-config"
sidebar_current: "docs-commands-operator-autopilot-get-config"
description: >
  Display the current Autopilot configuration.
---

# Command: operator autopilot get-config

The Autopilot operator command is used to view the current Autopilot configuration. See the
[Autopilot Guide](/guides/cluster/autopilot.html) for more information about Autopilot.

## Usage

```
nomad operator autopilot get-config [options]
```

## General Options

<%= partial "docs/commands/_general_options" %>

The output looks like this:

```
CleanupDeadServers = true
LastContactThreshold = 200ms
MaxTrailingLogs = 250
ServerStabilizationTime = 10s
RedundancyZoneTag = ""
DisableUpgradeMigration = false
UpgradeMigrationTag = ""
```

- `CleanupDeadServers` - Specifies automatic removal of dead
  server nodes periodically and whenever a new server is added to the cluster.

- `LastContactThreshold` - Specifies the maximum amount of
  time a server can go without contact from the leader before being considered
  unhealthy. Must be a duration value such as `10s`.

- `MaxTrailingLogs` - specifies the maximum number of log entries
  that a server can trail the leader by before being considered unhealthy.

- `ServerStabilizationTime` - Specifies the minimum amount of
  time a server must be stable in the 'healthy' state before being added to the
  cluster. Only takes effect if all servers are running Raft protocol version 3
  or higher. Must be a duration value such as `30s`.

- `RedundancyZoneTag` - Controls the node-meta key to use when
  Autopilot is separating servers into zones for redundancy. Only one server in
  each zone can be a voting member at one time. If left blank, this feature will
  be disabled.

- `DisableUpgradeMigration` - Disables Autopilot's upgrade
  migration strategy in Nomad Enterprise of waiting until enough
  newer-versioned servers have been added to the cluster before promoting any of
  them to voters.

- `UpgradeVersionTag` - Controls the node-meta key to use for
  version info when performing upgrade migrations. If left blank, the Nomad
  version will be used.
