{{#freestyle-usage "font-sizing"}}
  <div class="block">
    <h1 class="title">Large Title</h1>
    <p>Some prose to follow the large title. Not necessarily meant for reading.</p>
  </div>
  <div class="block">
    <h2 class="title is-4">Medium Title</h2>
    <p>Some prose to follow the large title. Not necessarily meant for reading.</p>
  </div>
  <div class="block">
    <h3 class="title is-5">Small Title</h3>
    <p>Some prose to follow the large title. Not necessarily meant for reading.</p>
  </div>
{{/freestyle-usage}}
