a {
  color: $blue;
  text-decoration: underline;

  &:hover {
    color: $blue;
    text-decoration: underline;
  }
}

code {
  color: currentColor;
  background: transparent;
  text-transform: none;
  padding: 0;
  font-size: 1em;
}

.nowrap {
  white-space: nowrap;
}

.is-interactive {
  cursor: pointer;
}
