---
layout: "docs"
page_title: "Commands: quota init"
sidebar_current: "docs-commands-quota-init"
description: >
  Generate an example quota specification.
---

# Command: quota init

The `quota init` command is used to create an example quota specification file
that can be used as a starting point to customize further.

~> Quota commands are new in Nomad 0.7 and are only available with Nomad
Enterprise.

## Usage

```
nomad quota init
```

## Init Options

* `-json`: Create an example JSON quota specification. 

## Examples

Create an example quota specification:

```
$ nomad quota init
Example quota specification written to spec.hcl
```
