<ol class="timeline">
  {{#if allocation.nextAllocation}}
    {{reschedule-event-row
      label="Next Allocation"
      allocation=allocation.nextAllocation
      time=allocation.nextAllocation.modifyTime}}
  {{/if}}
  {{#if allocation.hasStoppedRescheduling}}
    <li class="timeline-note" data-test-stop-warning>
      {{x-icon "warning" class="is-warning is-text"}} Nomad has stopped attempting to reschedule this allocation.
    </li>
  {{/if}}
  {{#if (and allocation.followUpEvaluation.waitUntil (not allocation.nextAllocation))}}
    <li class="timeline-note" data-test-attempt-notice>
      {{x-icon "clock" class="is-info is-text"}} Nomad will attempt to reschedule
      {{moment-from-now allocation.followUpEvaluation.waitUntil interval=1000}}
    </li>
  {{/if}}
  {{reschedule-event-row
    allocation=allocation
    linkToAllocation=false
    time=allocation.modifyTime}}

  {{#each (reverse allocation.rescheduleEvents) as |event|}}
    {{reschedule-event-row
      allocationId=event.previousAllocationId
      time=event.time}}
  {{/each}}
</ol>
