<table class="table is-striped is-fixed is-compact is-darkened">
  <thead>
    <tr>
      <th class="is-one-third">Name</th>
      <th>Value</th>
    </tr>
  </thead>
  <tbody>
    {{attributes-section attributes=attributes}}
  </tbody>
</table>
