{{#freestyle-usage "colors"}}
  {{freestyle-palette
      colorPalette=nomadTheme
      title="Nomad Theme"
      description="Accent and neutrals."}}

  {{freestyle-palette
      colorPalette=productColors
      title="Product Colors"
      description="Colors from other HashiCorp products. Often borrowed for alternative accents and color schemes."}}

  {{freestyle-palette
      colorPalette=emotiveColors
      title="Emotive Colors"
      description="Colors used in conjunction with an emotional response."}}
{{/freestyle-usage}}
