---
layout: "docs"
page_title: "Commands: operator autopilot set-config"
sidebar_current: "docs-commands-operator-autopilot-set-config"
description: >
  Modify the current Autopilot configuration.
---

# Command: operator autopilot set-config

The Autopilot operator command is used to set the current Autopilot configuration. See the
[Autopilot Guide](/guides/cluster/autopilot.html) for more information about Autopilot.


## Usage

```
nomad operator autopilot set-config [options]
```

## General Options

<%= partial "docs/commands/_general_options" %>

## Set Config Options

* `-cleanup-dead-servers` - Specifies whether to enable automatic removal of dead servers
upon the successful joining of new servers to the cluster. Must be one of `[true|false]`.

* `-last-contact-threshold` - Controls the maximum amount of time a server can go without contact
from the leader before being considered unhealthy. Must be a duration value such as `200ms`.

* `-max-trailing-logs` - Controls the maximum number of log entries that a server can trail
the leader by before being considered unhealthy.

* `-server-stabilization-time` - Controls the minimum amount of time a server must be stable in
the 'healthy' state before being added to the cluster. Only takes effect if all servers are
running Raft protocol version 3 or higher. Must be a duration value such as `10s`.

* `-disable-upgrade-migration` - (Enterprise-only) Controls whether Nomad will avoid promoting
new servers until it can perform a migration. Must be one of `[true|false]`.

* `-redundancy-zone-tag`- (Enterprise-only) Controls the
  [`redundancy_zone`](/docs/agent/configuration/server.html#redundancy_zone)
  used for separating servers into different redundancy zones.

* `-upgrade-version-tag` - (Enterprise-only) Controls the
  [`upgrade_version`](/docs/agent/configuration/server.html#upgrade_version) to
  use for version info when performing upgrade migrations. If left blank, the
  Nomad version will be used.

The output looks like this:

```
Configuration updated!
```

The return code will indicate success or failure.
