---
layout: "docs"
page_title: "Commands: acl bootstrap"
sidebar_current: "docs-commands-acl-bootstrap"
description: >
  The bootstrap command is used to create the initial ACL token.
---

# Command: acl bootstrap

The `acl bootstrap` command is used to bootstrap the initial ACL token.

## Usage

```
nomad acl bootstrap [options]
```

The `acl bootstrap` command requires no arguments.

## General Options

<%= partial "docs/commands/_general_options" %>

## Examples

Bootstrap the initial token:

```
$ nomad acl bootstrap
Accessor ID  = 5b7fd453-d3f7-6814-81dc-fcfe6daedea5
Secret ID    = 9184ec35-65d4-9258-61e3-0c066d0a45c5
Name         = Bootstrap Token
Type         = management
Global       = true
Policies     = n/a
Create Time  = 2017-09-11 17:38:10.999089612 +0000 UTC
Create Index = 7
Modify Index = 7
```
