<td data-test-server-name>{{#link-to "servers.server" agent.id class="is-primary"}}{{agent.name}}{{/link-to}}</td>
<td data-test-server-status>{{agent.status}}</td>
<td data-test-server-is-leader>{{if agent.isLeader "True" "False"}}</td>
<td data-test-server-address>{{agent.address}}</td>
<td data-test-server-port>{{agent.serfPort}}</td>
<td data-test-server-datacenter>{{agent.datacenter}}</td>
