---
layout: "docs"
page_title: "Commands: sentinel delete"
sidebar_current: "docs-commands-sentinel-delete"
description: >
  The sentinel delete command is used to delete a Sentinel policy.
---

# Command: sentinel delete

The `sentinel delete` command is used to delete a Sentinel policy.

~> Sentinel commands are new in Nomad 0.7 and are only available with Nomad
Enterprise.

## Usage

```
nomad sentinel delete [options] <Policy Name>
```

The `sentinel delete` command requires a single argument, the policy name.

## General Options

<%= partial "docs/commands/_general_options" %>

## Examples

Delete a policy:

```
$ nomad sentinel delete foo
Successfully deleted "foo" Sentinel policy!
```
