$consul-pink: #ff0087;
$consul-pink-dark: #c62a71;

$packer-blue: #1daeff;
$packer-blue-dark: #1d94dd;

$terraform-purple: #5c4ee5;
$terraform-purple-dark: #4040b2;

$vagrant-blue: #1563ff;
$vagrant-blue-dark: #104eb2;

$nomad-green: #25ba81;
$nomad-green-dark: #1d9467;
$nomad-green-darker: #16704d;
