{{#global-header class="page-header"}}
  {{#each breadcrumbs as |breadcrumb index|}}
    <li class="{{if (eq (inc index) breadcrumbs.length) "is-active"}}">
      {{#link-to
        data-test-breadcrumb=breadcrumb.label
        params=breadcrumb.args}}
        {{breadcrumb.label}}
      {{/link-to}}
    </li>
  {{/each}}
{{/global-header}}
{{#gutter-menu class="page-body"}}
  {{partial "allocations/allocation/task/subnav"}}
  <section class="section">
    {{task-log data-test-task-log allocation=model.allocation task=model.name}}
  </section>
{{/gutter-menu}}
