{{#each annotatedDeployments key="deployment.id" as |record|}}
  {{#if record.meta.showDate}}
    <li data-test-deployment-time class="timeline-note">
      {{#if record.deployment.version.submitTime}}
        {{moment-format record.deployment.version.submitTime "MMMM D, YYYY"}}
      {{else}}
        Unknown time
      {{/if}}
    </li>
  {{/if}}
  <li data-test-deployment class="timeline-object">
    {{job-deployment deployment=record.deployment}}
  </li>
{{/each}}
