<div class="loading-spinner">
  <div class="loading-spinner-square loading-spinner-square1"></div>
  <div class="loading-spinner-square loading-spinner-square2"></div>
  <div class="loading-spinner-square loading-spinner-square3"></div>
  <div class="loading-spinner-square loading-spinner-square4"></div>
  <div class="loading-spinner-square loading-spinner-square5"></div>
  <div class="loading-spinner-square loading-spinner-square6"></div>
  <div class="loading-spinner-square loading-spinner-square7"></div>
  <div class="loading-spinner-square loading-spinner-square8"></div>
  <div class="loading-spinner-square loading-spinner-square9"></div>
</div>
