{{#global-header class="page-header"}}
  {{#each breadcrumbs as |breadcrumb index|}}
    <li class="{{if (eq (inc index) breadcrumbs.length) "is-active"}}">
      {{#link-to data-test-breadcrumb=breadcrumb.label params=breadcrumb.args}}{{breadcrumb.label}}{{/link-to}}
    </li>
  {{/each}}
{{/global-header}}
{{#job-page/parts/body job=job onNamespaceChange=onNamespaceChange}}
  {{job-page/parts/error errorMessage=errorMessage onDismiss=(action "clearErrorMessage")}}

  {{job-page/parts/title job=job title=job.trimmedName handleError=(action "handleError")}}

  <div class="boxed-section job-stats">
    <div class="boxed-section-body">
      <span data-test-job-stat="type"><strong>Type:</strong> {{job.type}} | </span>
      <span data-test-job-stat="priority"><strong>Priority:</strong> {{job.priority}} </span>
      <span data-test-job-stat="parent">
        <strong>Parent:</strong>
        {{#link-to "jobs.job" job.parent (query-params jobNamespace=job.parent.namespace.name)}}
          {{job.parent.name}}
        {{/link-to}}
      </span>
      {{#if (and job.namespace system.shouldShowNamespaces)}}
        <span data-test-job-stat="namespace"> | <strong>Namespace:</strong> {{job.namespace.name}}</span>
      {{/if}}
    </div>
  </div>

  {{job-page/parts/summary job=job}}

  {{job-page/parts/placement-failures job=job}}

  {{job-page/parts/task-groups
    job=job
    sortProperty=sortProperty
    sortDescending=sortDescending
    gotoTaskGroup=gotoTaskGroup}}

  {{job-page/parts/evaluations job=job}}
{{/job-page/parts/body}}
