<div data-test-deployment-allocations class="boxed-section">
  <div class="boxed-section-head">
    Allocations
  </div>
  <div class="boxed-section-body is-full-bleed">
    {{#list-table
      source=deployment.allocations
      class="allocations" as |t|}}
      {{#t.head}}
        <th class="is-narrow"></th>
        <th>ID</th>
        <th>Modified</th>
        <th>Name</th>
        <th>Status</th>
        <th>Version</th>
        <th>Node</th>
        <th>CPU</th>
        <th>Memory</th>
      {{/t.head}}
      {{#t.body as |row|}}
        {{allocation-row data-test-deployment-allocation allocation=row.model context="job"}}
      {{/t.body}}
    {{/list-table}}
  </div>
</div>
