---
layout: "docs"
page_title: "Commands: namespace delete"
sidebar_current: "docs-commands-namespace-delete"
description: >
  The namespace delete command is used to delete a namespace.
---

# Command: namespace delete

The `namespace delete` command is used delete a namespace.

~> Namespace commands are new in Nomad 0.7 and are only available with Nomad
Enterprise.

## Usage

```
nomad namespace delete [options] <namespace>
```

The `namespace delete` command requires the name of the namespace to be deleted.

## General Options

<%= partial "docs/commands/_general_options" %>

## Examples

Delete a namespace

```
$ nomad namespace delete api-prod
Successfully deleted namespace "api-prod"!
```
