.status-text {
  font-weight: $weight-semibold;

  &.node-ready {
    color: $nomad-green-dark;
  }

  &.node-down {
    color: $danger;
  }

  &.node-initializing {
    color: $grey;
  }

  @each $name, $pair in $colors {
    $color: nth($pair, 1);
    $color-invert: nth($pair, 2);

    &.is-#{$name} {
      color: $color;
    }
  }
}
