{{#freestyle-collection inline=true as |collection|}}
  {{#collection.variant key="-apple-system"}}
    {{#freestyle-usage "font-apple-system" title="-apple-system"}}
      {{freestyle-typeface fontFamily="-apple-system"}}
    {{/freestyle-usage}}
  {{/collection.variant}}
  {{#collection.variant key="BlinkMacSystemFont"}}
    {{#freestyle-usage "font-blink-mac-system" title="BlinkMacSystemFont"}}
      {{freestyle-typeface fontFamily="BlinkMacSystemFont"}}
    {{/freestyle-usage}}
  {{/collection.variant}}
  {{#collection.variant key="Segoe UI"}}
    {{#freestyle-usage "font-segoe-ui" title="Segoe UI"}}
      {{freestyle-typeface fontFamily="Segoe UI"}}
    {{/freestyle-usage}}
  {{/collection.variant}}
  {{#collection.variant key="Roboto"}}
    {{#freestyle-usage "font-roboto" title="Roboto"}}
      {{freestyle-typeface fontFamily="Roboto"}}
    {{/freestyle-usage}}
  {{/collection.variant}}
  {{#collection.variant key="Oxygen Sans"}}
    {{#freestyle-usage "font-oxygen-sans" title="Oxygen Sans"}}
      {{freestyle-typeface fontFamily="Oxygen-Sans"}}
    {{/freestyle-usage}}
  {{/collection.variant}}
  {{#collection.variant key="Ubuntu"}}
    {{#freestyle-usage "font-ubuntu" title="Ubuntu"}}
      {{freestyle-typeface fontFamily="Ubuntu"}}
    {{/freestyle-usage}}
  {{/collection.variant}}
  {{#collection.variant key="Cantarell"}}
    {{#freestyle-usage "font-cantarell" title="Cantarell"}}
      {{freestyle-typeface fontFamily="Cantarell"}}
    {{/freestyle-usage}}
  {{/collection.variant}}
  {{#collection.variant key="Helvetica Neue"}}
    {{#freestyle-usage "font-helvetica-neue" title="Helvetica Neue"}}
      {{freestyle-typeface fontFamily="Helvetica Neue"}}
    {{/freestyle-usage}}
  {{/collection.variant}}
  {{#collection.variant key="sans-serif"}}
    {{#freestyle-usage "font-sans-serif" title="sans-serif"}}
      {{freestyle-typeface fontFamily="sans-serif"}}
    {{/freestyle-usage}}
  {{/collection.variant}}
  {{#collection.variant key="monospace"}}
    {{#freestyle-usage "font-monospace" title="monospace"}}
      {{freestyle-typeface fontFamily="monospace"}}
    {{/freestyle-usage}}
  {{/collection.variant}}
{{/freestyle-collection}}
