---
layout: "docs"
page_title: "Commands: namespace status"
sidebar_current: "docs-commands-namespace-status"
description: >
  The namespace status command is used to view the status of a particular
  namespace.
---

# Command: namespace status

The `namespace status` command is used to view the status of a particular
namespace.

~> Namespace commands are new in Nomad 0.7 and are only available with Nomad
Enterprise.

## Usage

```
nomad namespace status [options] <namespace_name>
```

## General Options

<%= partial "docs/commands/_general_options" %>

## Examples

View the status of a namespace:

```
$ nomad namespace status default
Name        = default
Description = Default shared namespace
Quota       = shared-default-quota

Quota Limits
Region  CPU Usage   Memory Usage
global  500 / 2500  256 / 2000
```
