.metric-group {
  display: inline-flex;
  width: auto;
  align-items: flex-start;

  .metric {
    padding: 0.75em 1em;
    border: 1px solid $grey-blue;
    text-align: center;
    display: flex;
    flex-direction: column;
    min-width: 120px;

    + .metric {
      margin-left: -1px;
    }

    &:first-child {
      border-top-left-radius: $radius;
      border-bottom-left-radius: $radius;
    }

    &:last-child {
      border-top-right-radius: $radius;
      border-bottom-right-radius: $radius;
    }

    @each $name, $pair in $colors {
      $color: nth($pair, 1);
      $color-invert: nth($pair, 2);

      &.is-#{$name} {
        border-color: $color;
        color: $color;
        background: lighten($color, 40%);

        .label {
          color: $color;
        }
      }
    }

    &.is-faded {
      color: $grey-blue;
      border-color: $grey-blue;
      background: $white;

      .label {
        color: $grey-blue;
      }
    }

    .label {
      font-size: 1.1em;
      font-weight: $weight-semibold;
      margin-bottom: 0;
    }

    .value {
      font-size: 2em;
      margin-bottom: 0;
    }
  }
}
