---
layout: "docs"
page_title: "Commands: quota status"
sidebar_current: "docs-commands-quota-status"
description: >
  The quota status command is used to view the status of a particular quota
  specification.
---

# Command: quota status

The `quota status` command is used to view the status of a particular quota
specification.

~> Quota commands are new in Nomad 0.7 and are only available with Nomad
Enterprise.

## Usage

```
nomad quota status [options] <quota_name>
```

## General Options

<%= partial "docs/commands/_general_options" %>

## Examples

View the status of a quota specification:

```
$ nomad quota status default-quota
Name        = default-quota
Description = Limit the shared default namespace
Limits      = 1

Quota Limits
Region  CPU Usage   Memory Usage
global  500 / 2500  256 / 2000
```
