---
layout: "inner"
page_title: "Community"
description: |-
  Nomad is an open source project with a growing community.
---

<h1>Community</h1>

<p>
  Nomad is an open source project with a growing community. There are
  active, dedicated users willing to help you through various mediums.
</p>
<p>
  <strong>Gitter:</strong> <a href="https://gitter.im/hashicorp-nomad/Lobby">Nomad Gitter Room</a>
</p>
<p>
  <strong>IRC:</strong> Use the <a href="https://irc.gitter.im">Gitter IRC bridge</a>
</p>
<p>
  <strong>Mailing list:</strong>
  <a href="https://groups.google.com/group/nomad-tool">Nomad Google Group</a>
</p>
<p>
  <strong>Bug Tracker:</strong>
  <a href="https://github.com/hashicorp/nomad/issues">Issue tracker
    on GitHub</a>. Please only use this for reporting bugs. Do not ask
  for general help here. Use IRC or the mailing list for that.
</p>

<h2>Community Tools</h2>
<p>
  These Nomad projects are created and managed by the amazing members of the
  Nomad community:
</p>

<ul>
  <li><a href="https://github.com/jippi/hashi-ui">hashi-ui</a> - Hashi-UI is a simple to deploy, web based UI for interacting with Nomad and Consul</li>
  <li><a href="https://github.com/iverberk/jenkins-nomad">nomad-jenkins</a> - This project uses Nomad to provision new Jenkins build slaves based on workload</li>
  <li><a href="https://github.com/elsevier-core-engineering/replicator">replicator</a> - Replicator is a fast and highly concurrent Go daemon that provides dynamic scaling of Nomad jobs and worker nodes</li>
  <li><a href="https://github.com/jrasell/levant">levant</a> - Levant is an open source templating and deployment tool that provides realtime feedback and detailed failure messages upon deployment issues</li>
</ul>

<em>
  Want to see your library here?
  <a href="https://github.com/hashicorp/nomad">Submit a Pull Request</a>.
  We also have a full list of <a href="/api/libraries-and-sdks.html">HTTP API Libraries</a>.
</em>
