---
layout: "docs"
page_title: "Commands: acl policy apply"
sidebar_current: "docs-commands-acl-policy-apply"
description: >
  The policy apply command is used to create or update ACL policies.
---

# Command: acl policy apply

The `acl policy apply` command is used to create or update ACL policies.

## Usage

```
nomad acl policy apply [options] <name> <path>
```

The `acl policy apply` command requires two arguments, the policy name and path to file.
The policy can be read from stdin by setting the path to "-".

## General Options

<%= partial "docs/commands/_general_options" %>

## Apply Options

* `-description`: Sets the human readable description for the ACL policy.

## Examples

Create a new ACL Policy:

```
$ nomad acl policy apply my-policy my-policy.json
Successfully wrote 'my-policy' ACL policy!
```
