{{#global-header class="page-header"}}
  {{#each breadcrumbs as |breadcrumb index|}}
    <li class="{{if (eq (inc index) breadcrumbs.length) "is-active"}}">
      {{#link-to params=breadcrumb.args}}{{breadcrumb.label}}{{/link-to}}
    </li>
  {{/each}}
{{/global-header}}
{{#gutter-menu class="page-body" onNamespaceChange=(action "gotoJobs")}}
  {{partial "jobs/job/subnav"}}
  <section class="section">
    {{job-deployments-stream deployments=deployments}}
  </section>
{{/gutter-menu}}
