.accordion {
  .accordion-head,
  .accordion-body {
    border: 1px solid $grey-blue;
    border-bottom: none;
    padding: 0.75em 1.5em;

    &:first-child {
      border-top-left-radius: $radius;
      border-top-right-radius: $radius;
    }

    &:last-child {
      border-bottom: 1px solid $grey-blue;
      border-bottom-left-radius: $radius;
      border-bottom-right-radius: $radius;
    }
  }

  .accordion-head {
    display: flex;
    background: $white-ter;
    flex: 1;

    &.is-light {
      background: $white;
    }

    &.is-inactive {
      color: $grey-light;
    }

    .accordion-head-content {
      width: 100%;
    }

    .accordion-toggle {
      flex-basis: 0;
      white-space: nowrap;
    }
  }
}
