---
layout: "docs"
page_title: "Commands: sentinel"
sidebar_current: "docs-commands-sentinel"
description: >
  The sentinel command is used to interact with Sentinel policies.
---

# Command: sentinel

The `sentinel` command is used to interact with Sentinel policies.

~> Sentinel commands are new in Nomad 0.7 and are only available with Nomad
Enterprise.

## Usage

Usage: `nomad sentinel <subcommand> [options]`

Run `nomad sentinel <subcommand> -h` for help on that subcommand. The following
subcommands are available:

* [`sentinel apply`][apply] - Create a new or update existing Sentinel policies
* [`sentinel delete`][delete] - Delete an existing Sentinel policies
* [`sentinel list`][list] - Display all Sentinel policies
* [`sentinel read`][read] - Inspects an existing Sentinel policies

[delete]: /docs/commands/sentinel/delete.html
[list]: /docs/commands/sentinel/list.html
[read]: /docs/commands/sentinel/read.html
[apply]: /docs/commands/sentinel/apply.html
