---
layout: "docs"
page_title: "Commands: status"
sidebar_current: "docs-commands-status"
description: >
  Display the status output for any Nomad resource.
---

# Command: status

The `status` command displays the status output for any Nomad resource.

## Usage

```
nomad status [options] <identifier>
```

The status command accepts any Nomad identifier or identifier prefix as its sole
argument. The command detects the type of the identifier and routes to the
appropriate status command to display more detailed output.

If the ID is omitted, the command lists out all of the existing jobs. This is
for backwards compatibility and should not be relied on.

## General Options

<%= partial "docs/commands/_general_options" %>

## Examples

Display the status of a job:

```
$ nomad status example
ID            = example
Name          = example
Submit Date   = 08/28/17 23:01:39 UTC
Type          = service
Priority      = 50
Datacenters   = dc1
Status        = running
Periodic      = false
Parameterized = false

Summary
Task Group  Queued  Starting  Running  Failed  Complete  Lost
cache       0       0         1        0       0         0

Latest Deployment
ID          = f5506391
Status      = running
Description = Deployment is running

Deployed
Task Group  Desired  Placed  Healthy  Unhealthy
cache       1        1       0        0

Allocations
ID        Node ID   Task Group  Version  Desired  Status   Created At
e1d14a39  f9dabe93  cache       0        run      running  08/28/17 23:01:39 UTC
```

Display the status of an allocation:

```
$ nomad status e1d14a39
ID                  = e1d14a39
Eval ID             = cc882755
Name                = example.cache[0]
Node ID             = f9dabe93
Job ID              = example
Job Version         = 0
Client Status       = running
Client Description  = <none>
Desired Status      = run
Desired Description = <none>
Created At          = 08/28/17 23:01:39 UTC
Deployment ID       = f5506391
Deployment Health   = healthy

Task "redis" is "running"
Task Resources
CPU        Memory           Disk     IOPS  Addresses
4/500 MHz  6.3 MiB/256 MiB  300 MiB  0     db: 127.0.0.1:21752

Task Events:
Started At     = 08/28/17 23:01:39 UTC
Finished At    = N/A
Total Restarts = 0
Last Restart   = N/A

Recent Events:
Time                   Type        Description
08/28/17 23:01:39 UTC  Started     Task started by client
08/28/17 23:01:39 UTC  Task Setup  Building Task Directory
08/28/17 23:01:39 UTC  Received    Task received by client
```

Display the status of a deployment:

```
$ nomad status f5506391
ID          = f5506391
Job ID      = example
Job Version = 0
Status      = successful
Description = Deployment completed successfully

Deployed
Task Group  Desired  Placed  Healthy  Unhealthy
cache       1        1       1        0
```

Display the status of a node:

```
$ nomad status f9dabe93
ID      = f9dabe93
Name    = nomad-server01
Class   = <none>
DC      = dc1
Drain   = false
Status  = ready
Drivers = docker,exec,java,qemu,raw_exec,rkt
Uptime  = 4h17m24s

Allocated Resources
CPU           Memory           Disk            IOPS
500/8709 MHz  256 MiB/2.0 GiB  300 MiB/24 GiB  0/0

Allocation Resource Utilization
CPU         Memory
3/8709 MHz  6.3 MiB/2.0 GiB

Host Resource Utilization
CPU           Memory           Disk
116/8709 MHz  335 MiB/2.0 GiB  12 GiB/38 GiB

Allocations
ID        Node ID   Task Group  Version  Desired  Status   Created At
e1d14a39  f9dabe93  cache       0        run      running  08/28/17 23:01:39 UTC
```
