---
layout: "docs"
page_title: "Commands: acl policy list"
sidebar_current: "docs-commands-acl-policy-list"
description: >
  The policy list command is used to list available ACL policies.
---

# Command: acl policy info

The `acl policy list` command is used to list available ACL policies.

## Usage

```
nomad acl policy list
```

## General Options

<%= partial "docs/commands/_general_options" %>
#
## List Options

* `-json` : Output the namespaces in their JSON format.

* `-t` : Format and display the namespaces using a Go template.

## Examples

List all ACL policies:

```
$ nomad acl policy list
Name        Description
default-ns  Write access to the default namespace
node-read   Node read access
```
