.job-diff {
  color: $grey-blue;
  font-family: $family-monospace;

  &.is-added,
  &.is-deleted {
    padding-left: 1ch;
  }

  &.is-edited {
    padding-left: 2ch;
  }

  .diff-section-label {
    text-indent: 0;
    padding-left: 2ch;

    &.is-added {
      text-indent: -2ch;
    }

    &.is-edited {
      text-indent: -4ch;
    }

    &.is-deleted {
      text-indent: -2ch;
    }

    .marker {
      &.is-added {
        color: $nomad-green;
      }

      &.is-edited {
        color: $orange;
      }

      &.is-deleted {
        color: $red;
      }
    }
  }

  .diff-section-block {
    padding-left: 0;
  }

  .diff-section-bold {
    color: $white;
    font-weight: $weight-bold;
  }

  .diff-section-change {
    color: $white;
    font-weight: $weight-bold;
  }

  .diff-section-table {
    display: table;
  }

  .diff-section-table-row {
    display: table-row;
  }

  .diff-section-table-cell {
    display: table-cell;

    & + .diff-section-table-cell {
      text-indent: 1ch;
    }
  }
}
