---
layout: "inner"
page_title: "Resources"
description: |-
  Nomad is widely deployed across a range of enterprises and business verticals.
---

<h1>Resources</h1>

<p>
  Nomad is a widely deployed open source project with a rapidly expanding 
  ecosystem. Numerous resources exist for learning, experimentation and 
  integration with third-party tools. 
</p>

<h2>Community</h2>

<p>
  <strong>Gitter:</strong> <a href="https://gitter.im/hashicorp-nomad/Lobby">Nomad Gitter Room</a>
</p>
<p>
  <strong>IRC:</strong> Use the <a href="https://irc.gitter.im">Gitter IRC bridge</a>
</p>
<p>
  <strong>Mailing list:</strong>
  <a href="https://groups.google.com/group/nomad-tool">Nomad Google Group</a>
</p>
<p>
  <strong>Bug Tracker:</strong>
  <a href="https://github.com/hashicorp/nomad/issues">Issue tracker
    on GitHub</a>. Please only use this for reporting bugs. Do not ask
  for general help here. Use IRC or the mailing list for that.
</p>

<h2>Videos</h2>

<ul>
  <li><a href="https://www.hashicorp.com/resources/jet-walmart-hashicorp-nomad-azure-run-apps">How <strong>Jet.com</strong> uses HashiCorp Nomad on Azure to run its applications</a></li>
  <li><a href="https://www.hashicorp.com/resources/end-to-end-production-nomad-citadel">End to end production Nomad at <strong>Citadel</strong></a></li>
  <li><a href="https://www.hashicorp.com/resources/nomad-overview-demo-citadel-use-case">HashiCorp Nomad demo and <strong>Citadel</strong> use case</a></li>
  <li><a href="https://www.youtube.com/watch?v=lTIvKZHedJQ">Nelson: Multi-region Container Orchestration for HashiCorp Nomad and Lyft Envoy</a></li>
  <li><a href="https://www.hashicorp.com/resources/elsevier-nomad-container-framework-demo"><strong>Elsevier's</strong> Container Framework with Nomad, Terraform, and Consul</a></li>
  <li><a href="https://www.hashicorp.com/resources/pagerduty-nomad-journey"><strong>PagerDuty's</strong> Nomadic Journey</a></li>
  <li><a href="https://www.hashicorp.com/resources/operating-jobs-scale-nomad">Operating jobs at scale with Nomad</a></li>
  <li><a href="https://www.hashicorp.com/resources/backend-batch-processing-nomad">Backend batch processing with Nomad</a></li>
</ul>

<h2>Blog Posts</h2>

<ul>
  <li><a href="https://medium.com/@henrikjohansen/cluster-scheduling-systems-for-large-scale-security-operations-351e7886852b">Cluster Scheduling Systems for Large Scale Security Operations</a> - Henrik Johansen</li>
  <li><a href="https://medium.com/@copyconstruct/schedulers-kubernetes-and-nomad-b0f2e14a896">Cluster Schedulers</a> - Cindy Sridharan</li>
  <li><a href="https://danielparker.me/nomad/hashicorp/schedulers/nomad/">Playing with Nomad from HashiCorp</a> - Daniel Parker, <strong>Target</strong></li>
  <li><a href="https://medium.com/@gabrielteles/migrating-from-aws-to-aws-e2bb1bfaa0e0">Migrating from AWS to AWS</a> - Gabriel Teles</li>
  <li><a href="https://stackshare.io/circleci/how-circleci-processes-4-5-million-builds-per-month">How <strong>CircleCI</strong> Processes 4.5 Million Builds Per Month</a> - Rob Zuber</li>
  <li><a href="http://timperrett.com/2017/05/13/nomad-with-envoy-and-consul">Envoy with Nomad and Consul</a> - Tim Perrett</li>
  <li><a href="https://www.hashicorp.com/blog/continuous-deployment-with-nomad-and-terraform">Continuous Deployment with Nomad and Terraform</a> - Nic Jackson</li>
  <li><a href="https://www.hashicorp.com/blog/running-apache-spark-on-nomad">Running Apache Spark on HashiCorp Nomad</a> - Rob Genova</li>
  <li><a href="https://www.hashicorp.com/blog/auto-bootstrapping-a-nomad-cluster">Auto-bootstrapping a Nomad Cluster</a> - Nic Jackson</li>
  <li><a href="https://www.hashicorp.com/blog/replacing-queues-with-nomad-dispatch">Replacing Queues with Nomad Dispatch</a> - Armon Dadgar</li>
  <li><a href="https://sanderknape.com/2016/08/nomad-consul-multi-datacenter-container-orchestration/">Multi-datacenter Container Orchestration with Nomad and Consul</a> - Sander Knape</li>
 <li><a href="https://www.hashicorp.com/blog/on-demand-container-storage-with-hashicorp-nomad/">On-demand Container Storage with HashiCorp Nomad</a> - Jeff Silberman, <strong>Portworx</strong></li>
</ul>

<h2>Tools for Provisioning and Experimentation</h2>

<ul>
  <li><a href="https://registry.terraform.io/search?q=nomad">Terraform Module Registry</a> - Official Nomad modules for AWS, Azure and Google Cloud</li>
  <li><a href="https://aws.amazon.com/quickstart/architecture/nomad/">HashiCorp Nomad on AWS</a> - AWS Quickstart Guide for Nomad</li>
  <li><a href="https://github.com/hashicorp/nomad/tree/master/terraform">Nomad Sandbox Environment for AWS and Azure</a> - Easily Provision a Combined Nomad, Consul and Vault Cluster on AWS or Azure</li>
  <li><a href="https://www.katacoda.com/hashicorp/scenarios/nomad-introduction">Introduction to Nomad (Katacoda)</a> - Interactive Introduction to Nomad</li>
  <li><a href="https://katacoda.com/hashicorp/scenarios/playground">Nomad Playground (Katacoda)</a> - Online Nomad Playground Environment</li>
</ul>

<h2>Integrations</h2>

<ul>
  <li><a href="https://apache.github.io/incubator-heron/docs/operators/deployment/schedulers/nomad/">Apache Heron Nomad Integration</a> - Realtime, Distributed Stream Processing on Nomad</li>
  <li><a href="https://github.com/jet/nomad-service-alerter"><strong>Jet.com</strong> Nomad Service Alerter</a> - Alerting for your Nomad services</li>
  <li><a href="https://getnelson.github.io/nelson/">Nelson</a> - Automated, Multi-region Container Deployment with HashiCorp Nomad</li>
  <li><a href="https://github.com/elsevier-core-engineering/replicator"><strong>Elsevier</strong> Replicator</a> - Automated Cluster and Job Scaling For HashiCorp Nomad</li>
  <li><a href="https://github.com/jrasell/levant">Levant</a> - An open source templating and deployment tool for HashiCorp Nomad jobs</li>
  <li><a href="https://github.com/underarmour/libra"><strong>Under Armour</strong> Libra</a> - A Nomad Auto Scaler</li>
  <li><a href="https://docs.datadoghq.com/integrations/nomad">Datadog Nomad Integration</a> - Monitor Nomad Clusters with Datadog</li>
  <li><a href="https://help.spotinst.com/hc/en-us/articles/115005038289-Nomad-Container-Management">Nomad Container Management with SpotInst</a> - Nomad Integration for SpotInst Elastigroup</li>
  <li><a href="https://github.com/jenkinsci/nomad-plugin">Jenkins Nomad Plugin</a> - Nomad Cloud Plugin for Jenkins</li>
  <li>OpenFaaS (<a href="https://registry.terraform.io/modules/nicholasjackson/open-faas-nomad/aws/0.4.0">Terraform</a> | <a href="https://github.com/openfaas/faas">GitHub</a>) - Nomad Integration for OpenFaas (Functions as a Service)</li>
  <li><a href="https://github.com/ValFadeev/rundeck-nomad-plugin">Rundeck Plugin for Nomad</a> - A Rundeck Plugin for Running Jobs on a Nomad Cluster</li>
</ul>

<h2>Other</h2>

<ul>
  <li><a href="https://github.com/jippi/hashi-ui">hashi-ui</a> - Hashi-UI is a simple to deploy, web based UI for interacting with Nomad and Consul</li>
  <li><a href="https://github.com/jippi/awesome-nomad">jippi/awesome-nomad</a> - A Curated List of Nomad Tools</li>
  <li><a href="https://github.com/jrasell/nomadfiles">jrasell/nomadfiles</a> - A Collection of Nomad Job Files for Deploying Applications</li>
</ul>

<h2>Trusted By</h2>

<ul>
  <li>Barclays</li>
  <li>Bownty.com</li>
  <li>CircleCI</li>
  <li>Citadel</li>
  <li>Deluxe Entertainment</li>
  <li>Elsevier</li>
  <li>Jet.com</li>
  <li>Pagerduty</li>
  <li>Pandora</li>
  <li>SAP Ariba</li>
  <li>SeatGeek</li>
  <li>Spaceflight Industries</li>
  <li>SpotInst</li>
  <li>Under Armour</li>
</ul>

<p>
<em>
  <strong>If you would like to have a resource or your company name added to this page, please</strong>  
  <a href="https://github.com/hashicorp/nomad">submit a Pull Request</a>.
</em>
</p>
